//
// Title:   CoTeXTool.h
// Created: Mon Mar 17 19:00:00 2003
// Authors:Tamy Boubekeur, Clment Bichel, Blanca B. Borro Escribano, Elisabeth Brunet
//
// Copyright (c) 2003, Funky CoTeX B Group, University of Bordeaux
//
// This file is part of the CoTeX plugin for PointShop 3D
// See http://www.cotex.fr.st for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact cotex@fr.st if any conditions of this
// licensing are not clear to you.
//

#ifndef __COTEXTOOL_H_
#define __COTEXTOOL_H_

// **********************************************
// To compile plugins (win32):
// 
// define TOOL_EXPORTS on the command line
//
// in Visual Studio:
// Project->Settings->C/C++, Category Preprocessor
// add TOOL_EXPORTS to 'preprocessor definitions'
// ***********************************************

#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qcursor.h>
#include <qevent.h>
#include <qwidget.h>
#include <qobject.h>

#include <ToolInterface.h>

#include "../ui/CoTeXConfigurationWidget.h"
#include "Interfaces/CoTeXTreeInterface.h"
#include "Interfaces/PrimitiveManagerInterface.h"
#include "Interfaces/OperatorManagerInterface.h"
#include "Interfaces/TextureInterface.h"

/// The Constructive Texturing Tool for PointShop
/**
 * This class implements a constructive texturing tool, 
 * allowing users to partition the euclidian space in 
 * different primitives.
 *  Differents types of primitives and textures are provided.
 * The Constructrive Texturing Tree can be save in a simple format (ctx),
 * for saving/restoring sessions, and exported in the HyperFun language for
 * being used in other tools (raytracers for example).
 *
 * @author  Tamy Boubekeur, Clment Bichel, Blanca B. Borro Escribano, Elisabeth Brunet
 * @version 1.0
 */
class CoTeXTool : public ToolInterface {

  Q_OBJECT

public:
	
  TOOL_API CoTeXTool ();
  TOOL_API virtual ~CoTeXTool ();

  // ***********************
  // interface ToolInterface
  // ***********************

  TOOL_API QToolButton* addButtonToToolBar (QToolBar *toolBar);
  TOOL_API QToolButton *getToolButton ();
  TOOL_API QCursor *getToolCursor () const;
  TOOL_API QWidget *getToolConfigurationWidget () const;
  TOOL_API bool isAvailable () const;
	
  TOOL_API void handleMousePressEvent   (QMouseEvent *me);
  TOOL_API void handleMouseReleaseEvent (QMouseEvent *me);
  TOOL_API void handleMouseMoveEvent    (QMouseEvent *me);
  TOOL_API void handleKeyPressEvent     (QKeyEvent   *ke);
  TOOL_API void handleKeyReleaseEvent   (QKeyEvent   *ke);

  TOOL_API void selectTool ();
  TOOL_API void deselectTool ();

public slots:

  void handleRendererWidgetRepainted ();
  
private slots:

// ******************
// Conections with UI
// ******************

  void handleExportHF (QString filename);
  void handleSave (QString filename);
  void handleLoad (QString filename);
  void handleReset ();
 
// *************************
// Conections with PointShop
// *************************

  void handleActiveObjectAboutToChange ();
  void handleActiveObjectChanged ();
  void handleObjectRemoved ();
   
private:

  QCursor                     *toolCursor;
  QToolButton                 *toolButton;
  CoTeXConfigurationWidget *cotexConfigurationWidget;
  
  TextureInterface::TextureMap textures; ///< created by the user
  CoTeXTreeInterface::CoTeXTreeMap coTeXTreeMap; ///< one colorCoTeXTree by Object
  PrimitiveManagerInterface::PrimitiveManagerMap primitiveManagerMap; ///< one by type of primitive
  PrimitiveManagerInterface::ObjectObjects objectObjects; ///< association "Object" <-> "Primitives objects for this object"
  OperatorManagerInterface::OperatorManagerMap operatorManagerMap; ///< one by type of operator
  
  PrimitiveInterface * lastPrimitiveCreated; ///< for easy and logic remove.
  Object * lastObjectTextured; ///< for easy and logic remove.
  Object * objectRemoved; ///< only used change<->remove communication.
  bool selected; ///< true if CoTeX is the active Tool

  /// True if CoTeX is avialable
  /**
   * The conditions for using CoTeX are : 
   *  - at least one object in the scene
   *  - the main renderer must be interactive (for vizualisation).
   */
  bool checkToolAvailability () const;

  /// make visible CoTeX 's Primtives Objects of an Object.
  void showCoTeXObjects (Object * object);
  
  /// make visible CoTeX 's Primtives Objects of the current Object if it has some.
  void showCoTeXObjects ();

  /// hide CoTeX 's Primtives Objects of an Object.
  void hideCoTeXObjects (Object * o);
  
  /// hide CoTeX 's Primtives Objects of the current  Object if it has some.
  void hideCoTeXObjects (); 
  
  /// Destroy CoTeX 's Primtives Objects of an Object.
  void destroyCoTeXObjects (Object * object);

  /// Remove CoTeX 's Primtives Objects of an Object from the scene.
  void removeCoTeXObjects (Object * object, bool autoDelete);

  ///  Export the CoTeXTree of the current object.
  bool exportHyperFun (QString fileName);

  /// Destroys all resources
  void cleanUp ();

};

// ****************
// access functions
// ****************

// extern "C" is needed so those functions can be accessed
// with GetProcAddress() (WIN32) and dlsym() (UNIX) respective
extern "C" {
  TOOL_API ToolInterface * createTool ();
  TOOL_API void destroyTool (ToolInterface * tool);
  TOOL_API const char * getToolName ();
}

#endif  // __COTEXTOOL_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
