#include <assert.h>
#include <math.h>

#include "Block.h"

static const double ALPHA = 1.0;
static const double BLOCK_BBOX_SCALE_FACTOR = 1.5;
static const double BLOCK_BBOX_THRESHOLD = 1.0;

Block::Block (QString n, const Vector3D& c, float dx, float dy, float dz, 
	      TextureInterface * t)
  : PrimitiveInterface (n, t), d (Vector3D (dx, dy, dz)), center (c) {
  assert (t != NULL && dx >= 0.0 && dy >= 0.0 && dz >= 0.0);
}

Block::Block(Block &block) : PrimitiveInterface (block.getName (),
						 block.getTexture ()), 
			     d (Vector3D (block.getDx (), block.getDy (), 
					  block.getDz ())),
			     center (block.getCenter ()) {}

void Block::setDx (float newDx) { 
  assert (newDx >= 0.0);
  d[0] = newDx;
}

void Block::setDy (float newDy) { 
  assert (newDy >= 0.0);
  d[1] = newDy;
}

void Block::setDz (float newDz) { 
  assert (newDz >= 0.0);
  d[2] = newDz;
}

BBoxMinMax Block::getBBoxMinMax () {
  BBoxMinMax bBoxMinMax (center[0] - d[0] - BLOCK_BBOX_THRESHOLD,
			 center[1] - d[1] - BLOCK_BBOX_THRESHOLD, 
			 center[2] - d[2] - BLOCK_BBOX_THRESHOLD,
			 center[0] + d[0] + BLOCK_BBOX_THRESHOLD,
			 center[1] + d[1] + BLOCK_BBOX_THRESHOLD,
			 center[2] + d[2] + BLOCK_BBOX_THRESHOLD);
  bBoxMinMax.scale (BLOCK_BBOX_SCALE_FACTOR); 
  return bBoxMinMax;
}

static inline  float min (float a, float b, float c) {
  if (a <= b && a <= c)
    return a;
  if (b <= a && b <=c)
    return b;
  return c;
}

double Block::eval (const Vector3D& position) {
  double f;
  Vector3D dist = (position - center) * (center + d - position);
    
  if(ALPHA == 1.0)
    f = min (dist[0], dist[1], dist[2]);
  else {
    f = dist[0]+dist[1]-sqrt(dist[0] * dist[0] + dist[1] * dist[1] 
			     -2.0 * ALPHA * dist[0] * dist[1]);
    f = f + dist[2] - sqrt(f * f + dist[2] * dist[2] - 2.0 * ALPHA * f * dist[2]);
  }
  
  return f;	
}


// **********************
// HyperFunable Interface
// **********************

void Block::writeHFString (QTextStream & outStream) {
  outStream << "my_model(x[3],a[1],s[3])" << endl << "{" << endl;
  writeDefArray (outStream);
  writeDefBody (outStream);
  outStream << endl << "}" << endl;
}


void Block::writeDefArray (QTextStream & outStream) {
  outStream << "array center_" << getName () << "[3];" << endl;
  outStream << "array color_" << getName () << "[3];" << endl << endl;
}


void Block::writeDefBody (QTextStream & outStream) {
  QString center_name = QString ("center_") + getName ();
  QString color_name; 

  if (getFather () != NULL) 
    color_name = "color_" + getName ();
  else 
    color_name = "s";
  outStream << center_name << " = [" << center[0] << ", " << center[1] << ", " 
	    << center[2] << "];" << endl;
  outStream << color_name << " = ";
  getTexture ()->writeDefBody (outStream);
  outStream << ";" << endl;  
  outStream << getName () << " = hfBlock (x," << center_name 
	    << ", " << d[0] << ", " << d[1] << ", " << d[2] << ");" 
	    << endl << endl;
}


// *****************
// Savable Interface
// *****************

void Block::save (QTextStream &stream) {
  stream << "primitive block " <<  getName () << " " << center[0] << " " << center[1] 
	 << " " << center[2] << " " << d[0] << " "   << d[1] << " " << d[2] 
	 << " " << getTexture ()->getName ();
}

void Block::load (QTextStream &stream) {
  QString n;
  float x, y, z, dx, dy, dz;
  stream >> n >> x >> y >> z >> dx >> dy >> dz;
  setName (n);
  center = Vector3D (x, y, z);
  d = Vector3D (dx, dy, dz);
}
