#include <assert.h>
#include <math.h>

#include "ConvPoint.h"

using namespace std;

static const double CONV_BBOX_THRESHOLD = 30.0; ///< if all the point of the skeleton are in the same plane.
static const double CONV_BBOX_SCALE = 2.0; ///< to ensure the whole convolution


ConvPoint::ConvPoint (QString n, double commonS, ConvSkeleton * ske, TextureInterface * tex,
		      double thres) : PrimitiveInterface (n, tex), 
				      S (commonS), squareS (S * S), threshold (thres),
				      skeleton (ske) {
  assert (tex != NULL && ske->size () >= 2 && thres > 0.0);
}

ConvPoint::ConvPoint(ConvPoint & conv) : PrimitiveInterface (conv.getName (),
							     conv.getTexture ()), 
				 S (conv.S), squareS (conv.S * conv.S), threshold (conv.getThreshold ()) {
  skeleton = new ConvSkeleton;
  unsigned int i;
  for (i = 0; i < conv.skeleton->size (); i++) {
    Vector3D tmp = (*conv.skeleton)[i];
    (*skeleton)[i] = Vector3D (tmp[0], tmp[1], tmp[2]); 
  }
}

ConvPoint::~ConvPoint () {
  skeleton->clear ();
  delete skeleton;
}

ConvPoint & ConvPoint::operator=(ConvPoint& conv) {
  S = conv.S;
  squareS = S * S;
  threshold = conv.threshold;
  skeleton = new ConvSkeleton;
  unsigned int i;
  for (i = 0; i < conv.skeleton->size (); i++) {
    Vector3D tmp = (*conv.skeleton)[i];
    (*skeleton)[i] = Vector3D (tmp[0], tmp[1], tmp[2]);
  }
  return (*this);
}

void ConvPoint::setS (float newS) { 
   S = newS;
   squareS = S * S;
}

void ConvPoint::setThreshold (double newThreshold) {
  assert (newThreshold > 0.0);
  threshold = newThreshold;
}

void ConvPoint::setSkeleton (ConvSkeleton * newSkeleton) { 
  assert (newSkeleton != NULL && skeleton->size () >= 2);
  skeleton = newSkeleton;
}

BBoxMinMax ConvPoint::getBBoxMinMax () {
  double xmin, ymin, zmin, xmax, ymax, zmax;
  
  xmin = xmax = skeleton->front ()[0];
  ymin = ymax = skeleton->front ()[1];
  zmin = zmax = skeleton->front ()[2];
  
  unsigned int k;
  for (k = 1; k < skeleton->size (); k++) {
    if ((*skeleton)[k][0] < xmin)
      xmin = (*skeleton)[k][0];
    if ((*skeleton)[k][1] < ymin)
      ymin = (*skeleton)[k][1];
    if ((*skeleton)[k][2] < zmin)
      zmin = (*skeleton)[k][2];
    if ((*skeleton)[k][0] > xmax)
      xmax = (*skeleton)[k][0];
    if ((*skeleton)[k][1] > ymax)
      ymax = (*skeleton)[k][1];
    if ((*skeleton)[k][2] > zmax)
      zmax = (*skeleton)[k][2];
  }
  
  BBoxMinMax bBoxMinMax(xmin - CONV_BBOX_THRESHOLD, ymin - CONV_BBOX_THRESHOLD, 
			zmin - CONV_BBOX_THRESHOLD, xmax + CONV_BBOX_THRESHOLD,
			ymax + CONV_BBOX_THRESHOLD, zmax + CONV_BBOX_THRESHOLD);
  bBoxMinMax.scale (CONV_BBOX_SCALE);
  
  return bBoxMinMax;
}

double ConvPoint::eval (const Vector3D& position) {
  return computePointValue (position);
}

inline double square (double x) { return x * x;}

double ConvPoint::computePointValue (const Vector3D & position) {
  double r2, f = 0.0;
  unsigned int i;
  unsigned int size = skeleton->size ();
  
  for(i = 0; i < size; i++) {
    Vector3D currentPoint = (*skeleton)[i]; 
    r2 = square (currentPoint[0] - position[0]) + square (currentPoint[1] - position[1]) 
      + square (currentPoint[2] - position[2]);
    f += 1 / square (1 + squareS * r2);
  }
  f = f - threshold;
  return f;
}

////////// TO OPTIMIZE ////////////////

double ConvPoint::computeCurveValue (const Vector3D& position) {
  Vector3D xt, xt2;
  int i, size;
  double f = 0.0;
  double ax, ay, az;
  double dx, dy, dz;
  double l, p, q, xx ;
  
  size = skeleton->size ();
  
  for (i = 0; i < size; i++) {
    xt = (*skeleton)[i];
    xt2 = xt * xt;
    
    l = sqrt (xt2[0] + xt2[1] + xt2[2]);
    
    assert (l != 0.0);
    
    ax = xt[0] / l;
    ay = xt[1] / l;
    az = xt[2] / l;
    
    dx = position[0] - xt[0];
    dy = position[1] - xt[1];
    dz = position[2] - xt[2];
    
    xx = dx * ax + dy * ay + dz * az;
    
    double squareS = S * S,
      squareDx = dx *dx,
      squareDy = dy *dy,
      squareDz = dz *dz,
      squareXx = xx * xx;
    
    p = sqrt (1 + squareS * (squareDx + squareDy + squareDz - squareXx));
    q = sqrt (1 + squareS * (squareDx + squareDy + squareDz + l * l - 2 * l * xx ));
    
    double squareP = p * p;
    
    f += xx / (2 * squareP * (squareP + squareS * squareXx)) + (l - xx) / (2 * squareP * q * q)
      + (atan (S * xx / p) + atan (S * (l - xx) / p)) / (2 * S * squareP * p);
  }
  
  f = f - threshold;
  return f;
}

void ConvPoint::writeHFString (QTextStream & outStream) {
  outStream << "my_model(x[3],a[1],s[3])" << endl << "{" << endl;
  writeDefArray (outStream);
  writeDefBody (outStream);
  outStream << endl << "}" << endl;
}


void ConvPoint::writeDefArray (QTextStream & outStream) {
  outStream << "array vect_" << getName () << "[" << (skeleton->size () * 3);
  outStream << "];" << endl;
  outStream << "array S_" << getName () << "[" << skeleton->size () << "];" << endl;
  outStream << "array color_" << getName () << "[3];" << endl << endl;
}


void ConvPoint::writeDefBody (QTextStream & outStream) {
  QString vect_name = QString ("vect_") + getName ();
  QString S_name = QString ("S_") + getName ();
  QString threshold_name = QString ("threshold_") + getName ();
  QString color_name; 
  unsigned int i, j;
  
  if (getFather () != NULL) 
    color_name = "color_" + getName ();
  else 
    color_name = "s";
  
  outStream << vect_name << " = [";
  for (i = 0 ; i < skeleton->size () - 1 ; i++) {
    for (j = 0; j < 3; j++)
      outStream << (*skeleton)[i][j] << ", ";
    outStream << endl;
  }
  outStream << (*skeleton)[i][0] << ", " << (*skeleton)[i][1] 
            << ", " << (*skeleton)[i][2] << "];" << endl; 
  
  outStream << S_name << " = [";
  for(i = 0 ; i < skeleton->size () - 1 ; i++)
    outStream << S << ", ";
  outStream << S << "];" << endl;
  outStream << threshold_name << " = " << threshold << ";" << endl ;
  outStream << color_name << " = ";
  getTexture ()->writeDefBody (outStream);
  outStream << ";" << endl;  
  outStream << getName () << " = hfConvPoint (x," << vect_name << "," << S_name;
  outStream << "," << threshold_name << ");" << endl << endl;
}

void ConvPoint::save (QTextStream &stream) {
  int size = skeleton->size ();
  stream << "primitive convpoint " << getName () << " " << S << " " << threshold << " " 
	 << size << " " ;
  for (int i = 0 ; i < size ; i++)
    for (int j = 0 ; j < 3 ; j++)
      stream << (*skeleton)[i][j] << " ";
  stream << getTexture ()->getName();
}

void ConvPoint::load (QTextStream &stream) {
  QString n;
  int size;
  float x, y, z;

  delete skeleton;
  skeleton = new ConvSkeleton();
  stream >> n >> S >> threshold >> size ;
  assert (size >= 2 && threshold > 0.0);
  setName(n);
  squareS = S * S;
  skeleton->resize (size);
  for (int i = 0 ; i < size ; i++) {
    stream >> x >> y >> z;
    (*skeleton)[i] = Vector3D (x, y, z);
  }
}
