#include "SurfelGenerator.h"

#include <math.h>
#include <typeinfo>
#include <assert.h>
#include <Scene.h>


#include "../Interfaces/PrimitiveManagerInterface.h"
#include "Polygonizer/hfpolymesh.h"

using namespace std;

// These constants defined the resolution of the point cloud generated
static const int ACCURACY_DEFAULT = 18;
static const int ACCURACY_SIZE = 3;

static PrimitiveInterface * currentPrimitive; ///< used for polygonization

// Function of evaluation for the polygonizer.
static double T (vector<double> X, double *) {
  return - currentPrimitive->eval (Vector3D (X[0], X[1], X[2]));
}

BBoxMinMax::BBoxMinMax (double minx, double miny, double minz,
			double maxx, double maxy, double maxz) {
  minMax.resize (6,0);
  minMax[0] = minx;
  minMax[1] = miny;
  minMax[2] = minz;
  minMax[3] = maxx;
  minMax[4] = maxy;
  minMax[5] = maxz;
}

void BBoxMinMax::scale (float coeff) {
  coeff -= 1.0;
  Vector3D min (minMax[0], minMax[1], minMax[2]);
  Vector3D max (minMax[3], minMax[4], minMax[5]);
  Vector3D center = (min + max) / 2;
  min += (min - center) * coeff;
  max += (max - center) * coeff;
  minMax[0] = min[0];
  minMax[1] = min[1];
  minMax[2] = min[2];
  minMax[3] = max[0];
  minMax[4] = max[1];
  minMax[5] = max[2];
}

static const double GENERAL_BASE_COEFF = 1000.0; ///< base for evaluation of the surfels's number

void generatePrimitiveSurfels (PrimitiveInterface * primitive, 
			       Object * primitiveObject, 
			       const BBoxMinMax & bBoxMinMax,
			       QColor surfelColor) {
  assert (primitive != NULL && primitiveObject != NULL);
  
  SurfelCollection * surfels = primitiveObject->getSurfelCollection ();
  vector<int> accuracy;
  HFPolyMesh * hfPolyMesh;

  currentPrimitive = primitive;
  
  try {
    hfPolyMesh = new HFPolyMesh (T);
  }
  catch (HFPError he) {
    cout << "CoTeX : Error of Surfel Generator: " << he.error << endl;
    cout << "Exiting..." << endl;
    exit (1);
  }
   
  accuracy.resize (ACCURACY_SIZE, ACCURACY_DEFAULT);
  
  // **************
  // Polygonization
  // **************

  hfPolyMesh->MinMax (bBoxMinMax.getMinMax ());
  hfPolyMesh->Grid (accuracy);
  hfPolyMesh->Normals (true);
  hfPolyMesh->Calc ();

  float surfelRadius = bBoxMinMax.getSize () / GENERAL_BASE_COEFF;
  int triangleNum = hfPolyMesh->TriangleNum ();
  Scene * scene = Scene::getInstance ();
  
  for (int triangle = 0; triangle < triangleNum; triangle++) {
    Vector3D position[3], normal[3];
    for (int i = 0; i < 3; i++) {
      int index = hfPolyMesh->itsData.getTriangle (triangle, i);
      position[i] = Vector3D (hfPolyMesh->itsData.getVertex (index, 0),
			      hfPolyMesh->itsData.getVertex (index, 1),
			      hfPolyMesh->itsData.getVertex (index, 2));
      normal[i] = Vector3D (- hfPolyMesh->itsData.getNormal (index, 0),
			    - hfPolyMesh->itsData.getNormal (index, 1),
			    - hfPolyMesh->itsData.getNormal (index, 2));
    }
    
    // ***********************************
    // Linear interpolation for extracting 
    // surfels of triangles set.
    // ***********************************
    
    Vector3D interpolatePosition = (position[0] + position[1] + position[2])/3;
    Vector3D interpolateNormal = (normal[0] + normal[1] + normal[2])/3;
    
    SurfelInterface * currentSurfel = surfels->addSurfel (false);
    scene->setSurfelPosition (currentSurfel, interpolatePosition, false);
    scene->setSurfelNormal (currentSurfel, interpolateNormal, false);
    scene->setSurfelRadius (currentSurfel, surfelRadius, false);
    scene->setSurfelDiffuseColor (currentSurfel, surfelColor.rgb (), false);
  }
  delete hfPolyMesh;
}
  

// ********************
// GENERATION OF SPHERE
// ********************

static const double SPHERE_BASE_COEFF = 50.0; ///< base for evaluation of the surfels's number

void generateSphereSurfels (Object * object,
			    float radius,
			    const ColorAttribute & surfelColor,
			    unsigned int meridien, 
			    unsigned int parallele) {
  assert (object != NULL && radius >= 0.0);

  SurfelCollection * surfels = object->getSurfelCollection ();
  unsigned int i, j;
  float phi;
  float theta = M_PI/2 - (M_PI / parallele);
  float surfelRadius ;

  // ***************************************
  // After tests, this value seem to be good
  // for having a propotionnal surfel radius 
  // to the sphere.
  // ***************************************

  
  surfelRadius = radius / SPHERE_BASE_COEFF;
  
    
  // *******************************************
  // Here, we don't use cartesian, but spherical 
  // coordinate for computing the sphere points 
  // by iteration.
  // *******************************************  
  
  for (i = 1; i < parallele; i++) {
    phi = 0;
    for (j = 0; j <= meridien; j++) {
      Scene * scene = Scene::getInstance ();
      Vector3D vect (sin (phi) * cos (theta), sin (theta), cos (phi) * cos (theta));
      SurfelInterface * currentSurfel = surfels->addSurfel (false);
      
      scene->setSurfelPosition (currentSurfel, vect * radius, false);
      scene->setSurfelNormal (currentSurfel, vect, false);
      scene->setSurfelRadius (currentSurfel, surfelRadius, false);
      scene->setSurfelDiffuseColor (currentSurfel, surfelColor.rgb (), false);
      
      phi += (2 * M_PI / meridien);
    }
    theta -= (M_PI / parallele);
  }
}



Object * generateObject (PrimitiveInterface * primitive,
			 Object * pointShopObject) {
  assert (primitive != NULL);
  Scene * scene = Scene::getInstance ();
  Object * object =
    new Object (scene->getUniqueName ("CoTeX" + primitive->getName ()));
  
  object->setPosition (pointShopObject->getPosition ());
  generatePrimitiveSurfels (primitive, object, primitive->getBBoxMinMax (),
			    createColorForViz (primitive->getTexture ()));
  return object;
}



ColorAttribute createColorForViz (TextureInterface * texture,
				  const Vector3D & position) {
  
  const AttributeInterface * baseAttribute = (*texture)(position);

  // there are some troubles with Visual C++ 6.0 with RTTI.
#ifndef _WIN32  
  if (typeid (*baseAttribute) == typeid (ColorAttribute))
    return ColorAttribute::getDifferentColor (*((ColorAttribute*)baseAttribute));
#endif
  return COTEX_SURFEL_COLOR;
}
  
